import logging
import os
import sqlite3
import sys
import time
from typing import List
import shutil
import typer
from skinny_orm.sqlite_orm import SqliteOrm

from .models import YoutubeVideo, VideoMp4, GcpClientSecret
from .utils import list_mp4_files_recursively, get_all_videos_mp4, create_dir_if_not_exists
from .youtube_helper import YoutubeHelper

CWD = os.getcwd()
TUBEDRIVE_PATH = os.path.join(CWD, '.tubedrive')
UPLOADED_PATH = os.path.join(TUBEDRIVE_PATH, '.uploaded')
create_dir_if_not_exists(TUBEDRIVE_PATH)
create_dir_if_not_exists(UPLOADED_PATH)
LIMIT_UPLOADS = 5
SLEEP_TIME = 10

logger: logging.Logger = logging.getLogger(__name__)
logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s %(name)s %(levelname)s - %(message)s',
    datefmt='%d-%m-%y %H:%M:%S',
    handlers=[logging.FileHandler(f'{TUBEDRIVE_PATH}/logs.log')]
)
logger.addHandler(logging.StreamHandler(sys.stdout))

app = typer.Typer()


@app.command()
def run(
        secret_path: str = typer.Option("", help="The path to your gcp client secret json file."),
        mock: bool = typer.Option(False, help="If true, mock the youtube upload to not use the api")
):
    logger.info(f"Running on directory: {CWD}")
    db_con = sqlite3.connect(f'{TUBEDRIVE_PATH}/tubedrive.db')
    orm = SqliteOrm(connection=db_con)

    gcp_client_secret = orm.select(GcpClientSecret).first()
    logger.info(f"Found gcp client secret path: {gcp_client_secret}")
    if gcp_client_secret is None:
        if not secret_path:
            secret_path = typer.prompt("What's the path of your gcp client secret json file ?")
        gcp_client_secret = GcpClientSecret(path=secret_path)
        orm.insert(gcp_client_secret)
    else:
        if secret_path:
            gcp_client_secret.path = secret_path
            logger.info(f"Updating gcp client secret: {gcp_client_secret}")
            orm.update(gcp_client_secret).using(GcpClientSecret.id)
    logger.info(f"gcp client secret: {gcp_client_secret}")

    video_uploaded_count = 0
    while True:

        if video_uploaded_count == LIMIT_UPLOADS:
            logger.info("Limit upload done!")
            break

        files = list_mp4_files_recursively(directory=CWD)
        all_videos_mp4: List[VideoMp4] = get_all_videos_mp4(files)[:1]  # Limiting to one video per run for now

        if len(all_videos_mp4) == 0:
            time.sleep(SLEEP_TIME)
            continue

        youtube_helper = YoutubeHelper(gcp_client_secret.path)
        for video_mp4 in all_videos_mp4:
            logger.info(f"Working with video {video_mp4.path}")
            youtube_video: YoutubeVideo = youtube_helper.upload_video(
                video_mp4.filename, f'{CWD}/{video_mp4.filename}', mock)
            video_uploaded_count += 1
            orm.insert(youtube_video)
            logger.info(f'{video_mp4.filename} uploaded => {youtube_video.url}')

            with open(video_mp4.link_path, mode="w") as f:
                f.write(f'<script>window.location = "{youtube_video.url}";</script>')

            logger.info(f"Move from {video_mp4.path} to {UPLOADED_PATH}/{video_mp4.filename}")
            shutil.move(video_mp4.path, f"{UPLOADED_PATH}/{video_mp4.filename}")

        orm.bulk_insert(all_videos_mp4)
        time.sleep(SLEEP_TIME)



import os
import pickle

import google_auth_oauthlib.flow
import googleapiclient
import googleapiclient.discovery
from googleapiclient.http import MediaFileUpload
from google.auth.transport.requests import Request
from .models import YoutubeVideo


class YoutubeHelper:
    YOUTUBE_UPLOAD_SCOPES = ["https://www.googleapis.com/auth/youtube.upload"]
    YOUTUBE_API_SERVICE_NAME = "youtube"
    YOUTUBE_API_VERSION = "v3"

    def __init__(self, gcp_client_secret_path: str, credentials_file: str = ".tubedrive/token.pickle"):
        # os.environ["OAUTHLIB_INSECURE_TRANSPORT"] = "1"
        # flow = google_auth_oauthlib.flow.InstalledAppFlow.from_client_secrets_file(
        #     gcp_client_secret_path, self.YOUTUBE_UPLOAD_SCOPES)
        # credentials = flow.run_local_server(port=8083)

        credentials = None
        # Check if the credentials file exists and load it
        if os.path.exists(credentials_file):
            with open(credentials_file, "rb") as token:
                credentials = pickle.load(token)

        # If there are no valid credentials available, then either refresh the token or log in.
        if not credentials or not credentials.valid:
            if credentials and credentials.expired and credentials.refresh_token:
                credentials.refresh(Request())
            else:
                flow = google_auth_oauthlib.flow.InstalledAppFlow.from_client_secrets_file(
                    gcp_client_secret_path, self.YOUTUBE_UPLOAD_SCOPES)
                credentials = flow.run_local_server(port=8083)
            # Save the credentials for the next run
            with open(credentials_file, "wb") as token:
                pickle.dump(credentials, token)

        self.youtube = googleapiclient.discovery.build(
            self.YOUTUBE_API_SERVICE_NAME, self.YOUTUBE_API_VERSION, credentials=credentials)

    def upload_video(self, video_name: str, video_path: str, mock: bool = False) -> YoutubeVideo:
        if mock:
            return YoutubeVideo(video_name, video_path, 'this is a mock url')
        request = self.youtube.videos().insert(
            part="snippet,status",
            body={
                "snippet": {
                    "categoryId": "22",
                    "description": "",
                    "title": video_name
                },
                "status": {
                    "privacyStatus": "private"
                }
            },
            media_body=MediaFileUpload(video_path)
        )
        response = request.execute()
        print(response)
        return YoutubeVideo(
            name=response["snippet"]["title"],
            path=video_path,
            url=f"vnd.youtube:{response['id']}"
        )



import os
from pathlib import Path
from typing import List

from moviepy.editor import VideoFileClip
from moviepy.video.compositing.concatenate import concatenate_videoclips

from .models import VideoMp4


def create_file_if_not_exists(path: str):
    fle = Path(path)
    fle.touch(exist_ok=True)


def create_dir_if_not_exists(path: str):
    if not os.path.exists(path):
        os.makedirs(path)


def list_mp4_files_recursively(directory) -> list:
    result = []
    for root, dirs, files in os.walk(directory, topdown=True):
        dirs[:] = [d for d in dirs if d not in ['.tubedrive']]
        for file in files:
            if file.endswith(".mp4"):
                result.append(os.path.join(root, file))
    return result


def get_all_videos_mp4(files: list) -> List[VideoMp4]:
    result = []
    end_time_s = 0
    for file_path in files:
        file_clip = VideoFileClip(file_path)
        start_time_s = end_time_s
        end_time_s += file_clip.duration
        video_mp4 = VideoMp4(
            path=file_path,
            length_s=file_clip.duration,
            start_time_s=int(start_time_s),
            end_time_s=int(end_time_s),
        )
        result.append(video_mp4)
    return result


def concatenate_videos(video_clip_paths, output_path, method="compose"):
    """Concatenates several video files into one video file
    and save it to `output_path`. Note that extension (mp4, etc.) must be added to `output_path`
    `method` can be either 'compose' or 'reduce':
        `reduce`: Reduce the quality of the video to the lowest quality on the list of `video_clip_paths`.
        `compose`: type help(concatenate_videoclips) for the info"""
    # create VideoFileClip object for each video file
    clips = [VideoFileClip(c) for c in video_clip_paths]
    if method == "reduce":
        # calculate minimum width & height across all clips
        min_height = min([c.h for c in clips])
        min_width = min([c.w for c in clips])
        # resize the videos to the minimum
        clips = [c.resize(newsize=(min_width, min_height)) for c in clips]
        # concatenate the final video
        final_clip = concatenate_videoclips(clips)
    elif method == "compose":
        # concatenate the final video with the compose method provided by moviepy
        final_clip = concatenate_videoclips(clips, method="compose")
    # write the output video file
    final_clip.write_videofile(output_path)



from dataclasses import dataclass, field
import uuid
from pathlib import Path


def get_new_uuid() -> str:
    return str(uuid.uuid4())


@dataclass
class GcpClientSecret:
    path: str
    id: str = field(default_factory=get_new_uuid)


@dataclass
class VideoMp4:
    path: str
    length_s: int
    start_time_s: int
    end_time_s: int
    id: str = field(default_factory=get_new_uuid)
    youtube_id: str = None

    @property
    def link_path(self) -> str:
        return self.path.replace('.mp4', '.html')

    @property
    def filename(self) -> str:
        return Path(self.path).name


@dataclass
class YoutubeVideo:
    name: str
    path: str
    url: str
    id: str = field(default_factory=get_new_uuid)

